package io.tswf.transmutations.view.generation.mapper

import groovy.transform.CompileStatic
import io.tswf.transmutations.view.exception.CantReachMethodException
import io.tswf.transmutations.view.exception.InvalidViewMethodException
import io.tswf.transmutations.view.generation.IViewMethodMapper
import io.tswf.transmutations.view.generation.ViewMethodMappingContext
import io.tswf.transmutations.view.util.ParametersHelper
import io.tswf.transmutations.view.util.ReturnTypeHelper
import org.objectweb.asm.commons.Method
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.MethodNode
import org.objectweb.asm.tree.VarInsnNode

import java.lang.reflect.Modifier

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class MethodInvokeMapper implements IViewMethodMapper {
    public static final Collection<String> RESERVED_NAMES = [ "<init>", "<clinit>" ]

    Method methodToInvokeDescription

    MethodInvokeMapper(Method methodToInvokeDescription) {
        this.methodToInvokeDescription = methodToInvokeDescription
    }

    @Override
    void handleViewMethodMapping(ViewMethodMappingContext context) {
        def targetClass = context.targetClass
        def methodToInvoke = targetClass.methods
                .find { it.name == methodToInvokeDescription.name && it.desc == methodToInvokeDescription.descriptor }
                .ensureNotNull { new CantReachMethodException(targetClass, methodToInvokeDescription) }

        def viewMethod = context.viewMethodMeta.viewMethod
        def newMethod = new MethodNode(ACC_PUBLIC, viewMethod.name, viewMethod.desc, null, new String[0])

        def viewMethodReturnType = viewMethod.returnType
        def methodToInvokerReturnType = methodToInvoke.returnType

        // Сравниваем параметры для вызываемого метода и view метода
        ParametersHelper.validateMethodCallParameters(methodToInvokeDescription, context.viewMethodMeta)

        if (methodToInvoke.name in RESERVED_NAMES) {
            throw new InvalidViewMethodException("name '${methodToInvoke.name}' is reserved and can not be used", context.viewMethodMeta)
        }

        // Сравниваем возвращаемые типы
        def needsToPopMethodReturn = ReturnTypeHelper.validateMethodReturnTypesAndCheckPop(methodToInvokeDescription, context.viewMethodMeta)

        def targetMethodIsStatic = Modifier.isStatic(methodToInvoke.access)

        // Для вызова нестатических методов загружаем this на стек
        if (!targetMethodIsStatic) {
            newMethod.instructions << new VarInsnNode(ALOAD, 0)
        }

        ParametersHelper.loadMethodCallParameters(newMethod.instructions, methodToInvokeDescription, context.viewMethodMeta)

        // Вызываем метод
        newMethod.instructions << new MethodInsnNode(targetMethodIsStatic ? INVOKESTATIC : INVOKEVIRTUAL, targetClass.name, methodToInvoke.name, methodToInvoke.desc, false)

        // Если метод-цель отдает значение, а у view метода отдаваемый тип - void, то отданное значение надо удалить со стека
        if (needsToPopMethodReturn) {
            newMethod.instructions << new InsnNode(methodToInvokerReturnType.size == 1 ? POP : POP2)
        }

        // Завершаем выполнение метода
        newMethod.instructions << new InsnNode(viewMethodReturnType.getOpcode(IRETURN))

        targetClass.methods << newMethod
    }
}
