package io.tswf.transmutations.view.reading.mappers

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.transmutations.common.accessors.AccessorHelper
import io.tswf.transmutations.common.reader.FieldDescriptionReader
import io.tswf.transmutations.common.remapping.TransmutationRemapper
import io.tswf.transmutations.view.annotation.FieldAccessor
import io.tswf.transmutations.view.exception.InvalidViewMethodException
import io.tswf.transmutations.view.generation.mapper.FieldAccessType
import io.tswf.transmutations.view.generation.mapper.FieldAccessorMapper
import io.tswf.transmutations.view.reading.IViewMethodMapperReader
import io.tswf.transmutations.view.reading.ViewMethodMapperReadingContext
import org.objectweb.asm.Type
import org.objectweb.asm.tree.AnnotationNode

@CompileStatic
@AutoService(IViewMethodMapperReader)
class FieldAccessorMapperReader implements IViewMethodMapperReader {

    public static final Collection<String> PREFIXES = AccessorHelper.GETTER_PREFIXES + AccessorHelper.SETTER_PREFIXES
    public static final String FIELD_DESCRIPTION_PARAMETER_NAME = "value"

    @Override
    void handleViewMethodReading(ViewMethodMapperReadingContext readingContext) {
        def viewMethod = readingContext.viewMethodMeta.viewMethod
        def viewMethodReturnType = viewMethod.returnType

        def fieldAccessorAnnotation = viewMethod.findAnnotationByType(FieldAccessor)
        def methodNameStartsFromAccessorPrefix = PREFIXES.any { prefix -> viewMethod.name.startsWith(prefix) && viewMethod.name.length() > prefix.length() }
        def methodIsFieldAccessor = fieldAccessorAnnotation != null || methodNameStartsFromAccessorPrefix

        if (methodIsFieldAccessor) {
            def parameters = viewMethod.parameterTypes

            // Проверяем параметры и возвращаемые типы методов
            if (parameters.size() > 1) {
                throw new InvalidViewMethodException("parameters count more than 1. Accessors may be getters (non-void methods with no parameters) and setters (void methods with single parameter)", readingContext.viewMethodMeta)
            }

            def isGetter = parameters.isEmpty()

            if (isGetter && viewMethodReturnType == Type.VOID_TYPE) {
                throw new InvalidViewMethodException("getter method can not return void, not a ${viewMethodReturnType.className}", readingContext.viewMethodMeta)
            }

            if (!isGetter && viewMethodReturnType != Type.VOID_TYPE) {
                throw new InvalidViewMethodException("setter method must be void type, not a ${viewMethodReturnType.className}", readingContext.viewMethodMeta)
            }

            // Вычитываем из аннотаций или из описаний методов имя и дескриптор целевого поля
            def rawFieldName, rawFieldType

            def fieldDescriptionAnnotation = fieldAccessorAnnotation?.findValueByName(FIELD_DESCRIPTION_PARAMETER_NAME, AnnotationNode)

            if (fieldDescriptionAnnotation != null) {
                (rawFieldName, rawFieldType) = FieldDescriptionReader.findFieldDescriptionFromAnnotation(fieldDescriptionAnnotation)
            } else {
                if (isGetter) {
                    rawFieldName = AccessorHelper.resolveFieldNameFromGetterMethod(readingContext.viewMethodMeta.viewMethod)
                    rawFieldType = viewMethod.returnType
                } else {
                    rawFieldName = AccessorHelper.resolveFieldNameFromSetterMethod(readingContext.viewMethodMeta.viewMethod)
                    rawFieldType = parameters.first()
                }
            }

            // Проверяем, что полученные дескриптор и имя валидны
            if (rawFieldName.isNullOrBlank()) {
                throw new InvalidViewMethodException("resolved target field type can not be empty or blank", readingContext.viewMethodMeta)
            }

            if (rawFieldType == null) {
                throw new InvalidViewMethodException("resolved target field type can not be empty or blank", readingContext.viewMethodMeta)
            }

            if (rawFieldType == Type.VOID_TYPE) {
                throw new InvalidViewMethodException("resolved target field type can not be void!", readingContext.viewMethodMeta)
            }

            // Проверяем валидность аксессора
            if (isGetter) {
                AccessorHelper.validateGetter(readingContext.viewMethodMeta.viewMethod)
            } else {
                AccessorHelper.validateSetter(readingContext.viewMethodMeta.viewMethod)
            }

            // Ремапим тип и имя поля
            def (mappedFieldName, mappedFieldType) = TransmutationRemapper.mapField(readingContext.viewMethodMeta.targetType, rawFieldName, rawFieldType)

            // Добавляем маппер в результат
            readingContext.result << new FieldAccessorMapper(mappedFieldName, mappedFieldType, isGetter ? FieldAccessType.GET : FieldAccessType.SET)
        }
    }
}
